"""
Create statistics for datasets.
"""
from collections import defaultdict

from clldutils.clilib import Table, add_format


def register(parser):
    add_format(parser, default='pipe')
    parser.add_argument(
        '--columns',
        help='list information on columns',
        action='store_true'
    )


def run(args):

    concepts = defaultdict(list)
    for i, ds in enumerate(args.api.datasets.values()):
        args.log.info('analyze ' + ds.id)
        for cid, concept in ds.concepts.items():
            if concept['concepticon_gloss']:
                concepts[cid, concept['concepticon_gloss']] += [ds.id]
    headers = ['No.', 'ID', 'Gloss', 'Datasets']
    with Table(args, *headers) as table:
        for i, ((cid, cgl), clists) in enumerate(
                sorted(concepts.items(), key=lambda x: (len(x[1]), x[0][0]))):
            table.append([i + 1, cid, cgl, len(clists)])



"""
Main command line interface to the pynorare package.
"""
import sys
import pathlib
import contextlib

from cldfcatalog import Config, Catalog
from clldutils.clilib import register_subcommands, get_parser_and_subparsers, ParserError, PathType
from clldutils.loglib import Logging
from pyconcepticon import Concepticon

from pynorare import NoRaRe
import pynorare.commands


def main(args=None, catch_all=False, parsed_args=None):
    try:  # pragma: no cover
        repos = Config.from_file().get_clone('concepticon')
    except KeyError:  # pragma: no cover
        repos = pathlib.Path('.')

    parser, subparsers = get_parser_and_subparsers('norare')
    parser.add_argument(
        '--repos',
        help="clone of concepticon/concepticon-data",
        default=repos,
        type=PathType(type='dir'))
    parser.add_argument(
        '--repos-version',
        help="version of repository data. Requires a git clone!",
        default=None)
    parser.add_argument(
        '--norarepo',
        default=pathlib.Path('.'),
        type=PathType(type='dir'))

    register_subcommands(subparsers, pynorare.commands)

    args = parsed_args or parser.parse_args(args=args)
    if not hasattr(args, "main"):  # pragma: no cover
        parser.print_help()
        return 1

    with contextlib.ExitStack() as stack:
        stack.enter_context(Logging(args.log, level=args.log_level))
        if args.repos_version:  # pragma: no cover
            # If a specific version of the data is to be used, we make
            # use of a Catalog as context manager:
            stack.enter_context(Catalog(args.repos, tag=args.repos_version))
        args.repos = Concepticon(args.repos)
        args.api = NoRaRe(args.norarepo, concepticon=args.repos)
        args.log.info('norare at {0}'.format(args.repos.repos))
        try:
            return args.main(args) or 0
        except KeyboardInterrupt:  # pragma: no cover
            return 0
        except ParserError as e:  # pragma: no cover
            print(e)
            return main([args._command, '-h'])
        except Exception as e:  # pragma: no cover
            if catch_all:  # pragma: no cover
                print(e)
                return 1
            raise


if __name__ == '__main__':  # pragma: no cover
    sys.exit(main() or 0)



"""
Download a given dataset.
"""
from pynorare.cli_util import add_datasets, iter_datasets


def register(parser):
    add_datasets(parser)


def run(args):
    for dataset in iter_datasets(args):
        dataset.download()



from pyconcepticon import Concepticon

from pynorare.dataset import get_dataset_cls


def add_datasets(parser):
    parser.add_argument(
        'dataset',
        nargs='+',
        help='select your dataset',
        type=str)


def iter_datasets(args):
    for dsid in args.dataset:
        cls = get_dataset_cls(args.api.datasets[dsid].path.parent)
        yield cls(repos=args.norarepo, concepticon=Concepticon(args.repos.repos))



import collections

from csvw.metadata import TableGroup
from csvw.dsv import reader
import attr
from cldfcatalog import Config
from pyconcepticon import Concepticon
from clldutils.source import Source
from clldutils.apilib import API
import pybtex.database

__all__ = ['NoRaRe']


@attr.s
class Column(object):
    dataset = attr.ib()
    name = attr.ib()
    structure = attr.ib()
    type = attr.ib()
    norare = attr.ib()
    rating = attr.ib()
    nameinsource = attr.ib()
    note = attr.ib()
    language = attr.ib()
    source = attr.ib()
    other = attr.ib()


@attr.s
class ConceptSetMeta(object):
    id = attr.ib()
    author = attr.ib()
    year = attr.ib()
    tags = attr.ib(
        repr=False,
        converter=lambda x: [y.strip() for y in x.split(',')] if isinstance(x, str) else x)
    source_language = attr.ib(repr=False)
    target_language = attr.ib(repr=False)
    url = attr.ib(repr=False)
    refs = attr.ib(repr=False)
    note = attr.ib(repr=False)
    alias = attr.ib(repr=False)
    norare = attr.ib(default=None, repr=False)
    path = attr.ib(default=None, repr=False)

    @property
    def table(self):
        return TableGroup.from_file(self.path).tabledict[self.id + '.tsv']

    @property
    def concepts(self):
        concepts = collections.OrderedDict()
        for row in self.table:
            concepts[row['CONCEPTICON_ID']] = collections.OrderedDict(
                [(k.lower(), v) for k, v in row.items()])
        return concepts

    @property
    def columns(self):
        columns = collections.OrderedDict(
            [(c.name.lower(), {'nameinsource': str(c.titles) if c.titles else ''})
             for c in self.table.tableSchema.columns])

        for c in self.table.tableSchema.columns:
            columns[c.name.lower()] = Column(
                dataset=self.id,
                name=c.name.lower(),
                structure=self.norare.annotations[self.id].get(
                    c.name.lower(), {'structure': ''})['structure'],
                type=self.norare.annotations[self.id].get(
                    c.name.lower(), {'type': ''})['type'],
                norare=self.norare.annotations[self.id].get(
                    c.name.lower(), {'norare': ''})['norare'],
                rating=self.norare.annotations[self.id].get(
                    c.name.lower(), {'rating': ''})['rating'],
                note=self.norare.annotations[self.id].get(
                    c.name.lower(), {'note': ''})['note'],
                source=self.norare.annotations[self.id].get(
                    c.name.lower(), {'source': ''})['source'] or self.refs,
                language=self.norare.annotations[self.id].get(
                    c.name.lower(), {'language': ''})['language'],
                nameinsource=str(c.titles) if c.titles else '',
                other=self.norare.annotations[self.id].get(
                    c.name.lower(), {'other': ''})['other'])

        return columns


class NoRaRe(API):
    """
    Basic class for handling the norms-rates-relations data.
    """
    def __init__(self, repos=None, datasets=None, concepticon=None):
        API.__init__(self, repos)
        self.datasets = datasets or collections.OrderedDict()

        concepticon = concepticon
        if not concepticon:  # pragma: no cover
            try:
                concepticon = Concepticon(Config.from_file().get_clone('concepticon'))
            except KeyError:
                pass

        datasets = set()
        self.annotations = collections.defaultdict(lambda: collections.OrderedDict())
        for row in reader(self.repos / 'norare.tsv', delimiter='\t', dicts=True):
            self.annotations[row['DATASET']][row['NAME'].lower()] = {
                k.lower(): row[k] for k in [
                    'DATASET', 'NAME',
                    'LANGUAGE', 'STRUCTURE',
                    'TYPE', 'NORARE', 'RATING', 'SOURCE', 'OTHER', 'NOTE']}
            datasets.add(row['DATASET'])

        # get bibliography
        self.refs = collections.OrderedDict()
        with self.repos.joinpath('references', 'references.bib').open(encoding='utf-8') as fp:
            for key, entry in pybtex.database.parse_string(
                    fp.read(), bib_format='bibtex').entries.items():
                self.refs[key] = Source.from_entry(key, entry)

        all_refs = set(self.refs)
        if concepticon:
            all_refs = all_refs.union(concepticon.bibliography)

        for row in reader(self.repos / 'concept_set_meta.tsv', delimiter='\t', dicts=True):
            row['norare'] = self
            row['path'] = self.repos.joinpath(
                'concept_set_meta', row['ID'], row['ID'] + '.tsv-metadata.json')
            self.datasets[row['ID']] = ConceptSetMeta(**{k.lower(): v for k, v in row.items()})
            self.datasets[row['ID']].source_language = [
                l.lower().strip() for l in self.datasets[row['ID']].source_language.split(',')]

        # remaining datasets come from concepticon, we identify them from datasets
        concepticon_datasets = [d for d in datasets if d not in self.datasets]
        for dataset in concepticon_datasets:
            ds = concepticon.conceptlists[dataset]
            self.datasets[ds.id] = ConceptSetMeta(
                id=ds.id,
                author=ds.author,
                year=ds.year,
                tags=', '.join(ds.tags),
                source_language=ds.source_language,
                target_language=ds.target_language,
                url=ds.url,
                refs=ds.refs,
                note=ds.note,
                alias=ds.alias,
                norare=self,
                path=concepticon.repos.joinpath(
                    'concepticondata', 'conceptlists', ds.id + '.tsv-metadata.json')
            )

        for dataset in self.datasets.values():
            if dataset.refs:
                refs = [dataset.refs] if isinstance(dataset.refs, str) else dataset.refs
                for ref in refs:
                    if ref not in all_refs:  # pragma: no cover
                        raise ValueError('missing references.bib: {}'.format(ref))



"""
Check the norm data TSV file.
"""


def run(args):
    for i, ds in enumerate(args.api.datasets.values()):
        visited = set()
        args.log.info('checking {0}'.format(ds.id))
        for colid, column in ds.columns.items():
            if colid in args.api.annotations[ds.id]:
                uniq = '-'.join([
                    column.language,
                    column.norare,
                    column.type,
                    column.structure,
                    column.other,
                ])
                if uniq in visited:  # pragma: no cover
                    args.log.warn('non-unique value {0} in {1} / {2}'.format(uniq, ds.id, colid))
                else:
                    visited.add(uniq)



import zipfile
import collections
from urllib.request import urlretrieve

from cldfcatalog import Config
from csvw.dsv import reader
from pyconcepticon import Concepticon

import xlrd


def get_mappings(concepticon=None):
    concepticon = concepticon or Concepticon(Config.from_file().get_clone('concepticon'))
    paths = {p.stem.split('-')[1]: p
             for p in concepticon.repos.joinpath('mappings').glob('map-*.tsv')}
    mappings = {}
    for language, path in paths.items():
        mappings[language] = collections.defaultdict(set)
        for line in reader(path, delimiter='\t', dicts=True):
            gloss = line['GLOSS'].split('///')[1]
            oc = concepticon.conceptsets[line['ID']].ontological_category
            mappings[language][gloss].add((line['ID'], int(line['PRIORITY']), oc))
    for language, path in paths.items():
        for k, v in mappings[language].items():
            # We sort concepticon matches for a given gloss by descending priority and ascending
            # Concepticon ID.
            mappings[language][k] = sorted(
                v, key=lambda x: (x[1], -int(x[0])), reverse=True)
    return mappings, concepticon


def get_excel(path, sheet_index, dicts=False):
    xlfile = xlrd.open_workbook(str(path))
    sheet = xlfile.sheet_by_index(sheet_index)
    sheet = [sheet.row_values(i) for i in range(0, sheet.nrows)]
    return [dict(zip(sheet[0], row)) for row in sheet[1:]] if dicts else sheet


def download_archive(url, target, filename, path, cls=zipfile.ZipFile):
    urlretrieve(url, str(target))
    with open(str(target), 'rb') as f:
        with cls(f) as archive:
            archive.extract(filename, path=str(path))



"""
Create data for the web-application.
"""
import json
import collections


def run(args):
    meta, data = {}, collections.defaultdict(dict)
    for i, ds in enumerate(args.api.datasets.values()):
        args.log.info('analyze ' + ds.id)
        meta[ds.id] = {
            'author': ds.author,
            'year': ds.year,
            'tags': ds.tags,
            'source_languages': ds.source_language,
            'target_languages': ds.target_language,
        }
        for cid, concept_ in ds.concepts.items():
            concept = {}
            for colid, column in ds.columns.items():
                if colid in args.api.annotations[ds.id]:
                    concept[colid] = {
                        'value': concept_[colid],
                        'language': column.language,
                        'norare': column.norare,
                        'type': column.type,
                        'other': column.other,
                        'structure': column.structure,
                    }
            if concept_['concepticon_gloss']:
                data[str(cid) + '/' + concept_['concepticon_gloss']][ds.id] = concept

    args.norarepo.joinpath('app', 'data').mkdir(parents=True, exist_ok=True)
    with args.norarepo.joinpath('app', 'data', 'norare.js').open('w', encoding='utf8') as f:
        f.write('var NRRDATA = ' + json.dumps(data) + ';\n')
        f.write('var NRRMETA = ' + json.dumps(meta) + ';\n')



import inspect
import pathlib
import importlib.machinery
import collections
from urllib.request import urlretrieve

from csvw.metadata import TableGroup
from csvw.dsv import reader

from pynorare.files import get_mappings, get_excel, download_archive
from pynorare.log import get_logger

__all__ = ['NormDataSet', 'get_dataset_cls']


def get_dataset_cls(dsdir, dsid=None):
    dsid = dsid or dsdir.name
    mod = importlib.machinery.SourceFileLoader(
        'norare.{}'.format(dsid.replace('-', '_')), str(dsdir / 'map.py')).load_module()
    for _, cls in inspect.getmembers(mod, inspect.isclass):
        print(_, cls)
        if issubclass(cls, NormDataSet):
            return cls


class NormDataSet:
    id = ""

    def __init__(self, repos=pathlib.Path('.'), concepticon=None, mappings=None):
        self.repos = repos
        self.mapped = collections.defaultdict(list)
        if not mappings:  # pragma: no cover
            mappings, concepticon = get_mappings(concepticon)
        self.mappings, self.concepticon = mappings, concepticon
        self.dir = self.repos / 'concept_set_meta' / self.id
        self.raw_dir = self.dir / 'raw'
        if not self.raw_dir.exists():
            self.raw_dir.mkdir()
        self.fname = self.id + '.tsv'
        self.log = get_logger()

    @property
    def tablegroup(self):
        return TableGroup.from_file(self.dir.joinpath(self.fname + '-metadata.json'))

    @property
    def table(self):
        return self.tablegroup.tabledict[self.fname]

    @property
    def columns(self):
        return self.table.tableSchema.columns

    def validate(self):
        mappings = list(self.table)
        if mappings:
            self.log.info('metadata file can be loaded')

        if 'CONCEPTICON_ID' in mappings[0] and \
                'CONCEPTICON_GLOSS' in mappings[0] and \
                'LINE_IN_SOURCE' in mappings[0]:
            self.log.info('concepticon data present in data')

    def map(self):  # pragma: no cover
        self.log.warning("Function MAP is not defined")

    def download(self):  # pragma: no cover
        self.log.warning("Function DOWNLOAD is not defined")

    def download_zip(self, url, target, filename):
        download_archive(url, self.raw_dir.joinpath(target), filename, self.raw_dir)
        self.log.info('Downloaded {0} successfully.'.format(url))

    def download_file(self, url, target):
        urlretrieve(url, str(self.raw_dir / target))
        self.log.info('Downloaded {0} successfully.'.format(url))

    def get_csv(self, path, delimiter="\t", dicts=True, coding="utf-8"):
        self.log.info('load data from {0}'.format(path))
        return list(reader(self.raw_dir / path, delimiter=delimiter, dicts=dicts, encoding=coding))

    def get_excel(self, path, sidx, dicts=True):
        sheet = get_excel(self.raw_dir.joinpath(path), sidx, dicts)
        self.log.info('load data from {0}'.format(path))
        return sheet

    def extract_data(self,
                     dicts,
                     gloss='ENGLISH',
                     language='en',
                     pos=False,
                     pos_mapper=False,
                     pos_name='POS'):
        pos_mapper = pos_mapper or {}

        rename = {str(c.titles): c.name for c in self.columns if c.titles}
        # (conceptset ID, list of rows with matching glosses)
        mapped = collections.defaultdict(list)

        for i, row in enumerate(dicts, start=1):
            new_row = {rename.get(k, k): v for k, v in row.items()}
            new_row.setdefault('LINE_IN_SOURCE', i)
            mappings = self.mappings[language].get(new_row[gloss])
            if mappings:
                match, priority = None, None
                if pos:
                    for match, priority, pos_ in reversed(mappings):
                        if pos_ == pos_mapper.get(new_row[pos_name], new_row[pos_name]) and match:
                            break
                else:
                    match, priority, pos_ = mappings[0]

                if match:
                    new_row['CONCEPTICON_ID'] = str(match)
                    new_row['CONCEPTICON_GLOSS'] = self.concepticon.conceptsets[match].gloss
                    new_row['_PRIORITY'] = priority
                    mapped[match].append(new_row)

        table = []
        for key, rows in sorted(mapped.items(), key=lambda x: x[0]):
            # We choose one representative gloss in the raw data for each conceptset ID, selecting
            # by higher priority and lower line number in the raw data.
            table.append(sorted(rows, key=lambda x: (x['_PRIORITY'], x['LINE_IN_SOURCE']))[0])

        self.mapped = mapped
        self.table.write(table, base=self.dir)



"""
List all datasets.
"""
import collections

from clldutils.clilib import Table, add_format

from pynorare.util import progressbar


def register(parser):
    add_format(parser, default='pipe')
    parser.add_argument(
        '--columns',
        help='list information on columns',
        action='store_true'
    )


def run(args):
    concepts = set()
    columns = collections.defaultdict(list)

    headers = ['No', 'Dataset', 'Field', 'Ln', 'Norare', 'Structure', 'Type'] \
        if args.columns else ['ID', 'Author', 'Year', 'Languages', 'Tags', 'Ratings', 'Concepts']
    with Table(args, *headers) as table:
        for i, ds in progressbar(enumerate(args.api.datasets.values())):
            if not args.columns:
                table.append([
                    ds.id,
                    ds.author.replace(' AND ', ' and ').split(' and ')[0],
                    ds.year,
                    ', '.join(ds.source_language[:3]),
                    ', '.join(ds.tags),
                    len(ds.columns) - 3,
                    len(ds.concepts)
                ])
                concepts.update(ds.concepts)
            else:
                for column in ds.columns:
                    if column not in [
                        'concepticon_id',
                        'concepticon_gloss',
                        'line_in_source',
                        'english',
                        'german',
                        'polish',
                        'spanish',
                        'chinese',
                        'french',
                        'dutch',
                    ]:
                        columns[(ds.id, column)] += [(
                            ds.columns[column].language,
                            ds.columns[column].norare,
                            ds.columns[column].structure,
                            ds.columns[column].type,
                        )]
        if not args.columns:
            table.append([
                '-',
                'TOTAL',
                '-',
                '-',
                '-',
                sum([x[-2] for x in table]),
                len(concepts)])
        else:
            for i, (k, v) in enumerate(sorted(columns.items(), key=lambda x: (x[0][1], x[0][0]))):
                table.append((
                    i + 1,
                    k[0],
                    k[1],
                    ', '.join(list(set([x[0] for x in v])))[:30],
                    ', '.join(list(set([x[1] for x in v])))[:15],
                    ', '.join(list(set([x[2] for x in v])))[:15],
                    ', '.join(list(set([x[3] for x in v])))[:15],
                ))
